from random import Random
from itertools import repeat
from numbers import Real
from typing import Any, Callable, TypeVar, Optional, Iterator, cast, List, Mapping, Union

from greensim import happens


T = TypeVar("T")
VarRandom = Iterator[T]
RandomOpt = Optional[Random]


_default_random: Random = Random()


def set_default_random(rng: Random) -> None:
    global _default_random
    _default_random = rng


def _get_default_random() -> Random:
    """For testing purposes only."""
    return _default_random


def _ordef(rng: RandomOpt) -> Random:
    return rng or _default_random


def constant(value: T) -> VarRandom[T]:
    yield from repeat(value)


def _vr_from_fn(fn: Callable[..., T], *args: Any, **kwargs: Any) -> VarRandom[T]:
    while True:
        yield fn(*args, **kwargs)


def linear(vr: VarRandom[Real], slope: Real, shift: Real) -> VarRandom[Real]:
    yield from map(lambda x: slope * x + shift, vr)


def bounded(gen: VarRandom[Real], lower: Optional[Real] = None, upper: Optional[Real] = None) -> VarRandom[Real]:
    for x in gen:
        if lower is not None:
            x = max(lower, x)
        if upper is not None:
            x = min(x, upper)
        yield x


def project_int(vr: VarRandom[Real]) -> VarRandom[int]:
    yield from map(cast(Callable[[Real], int], int), vr)


def uniform(lower: Real, upper: Real, rng: RandomOpt = None) -> VarRandom[float]:
    yield from _vr_from_fn(_ordef(rng).uniform, lower, upper)


def expo(mean: Real, rng: RandomOpt = None) -> VarRandom[float]:
    yield from _vr_from_fn(_ordef(rng).expovariate, 1.0 / mean)


def normal(mean: Real, std_dev: Real, rng: RandomOpt = None) -> VarRandom[float]:
    yield from _vr_from_fn(_ordef(rng).normalvariate, mean, std_dev)


def poisson_process(mean_rate: Real, rng: RandomOpt = None) -> Callable[..., Any]:
    return happens(expo(1.0 / mean_rate, rng))


def distribution(distr: Union[List[T], Mapping[T, Real]], rng: RandomOpt = None) -> VarRandom[T]:
    if isinstance(distr, Mapping):
        vw = list(distr.items())
        values = [v for v, _ in vw]
        weights = [w for _, w in vw]
    else:
        values = distr
        weights = cast(List[Real], [1.0 for v in values])
    yield from map(lambda x: cast(T, x[0]), _vr_from_fn(_ordef(rng).choices, values, weights=weights))



"""
Progress tracking tools for simulations.
"""

from math import ceil, inf
import sys
import time
from typing import Callable, Sequence, Tuple, IO, Optional, Union

from greensim import now, advance, stop


MetricProgress = Sequence[float]
MeasureProgress = Callable[[], MetricProgress]
MeasureComparison = Sequence[Tuple[float, float]]
CaptureProgress = Callable[[float, float, MeasureComparison], None]


def combine(*measures: MeasureProgress) -> MetricProgress:
    """Combines multiple progress measures into one metric."""
    return sum((list(measure()) for measure in measures), [])


def sim_time():
    return [now()]


def capture_print(file_dest_maybe: Optional[IO] = None):
    """Progress capture that writes updated metrics to an interactive terminal."""
    file_dest: IO = file_dest_maybe or sys.stderr

    def _print_progress(progress_min: float, rt_remaining: float, _mc: MeasureComparison) -> None:
        nonlocal file_dest
        percent_progress = progress_min * 100.0
        time_remaining, unit = _display_time(rt_remaining)
        print(
            f"Progress: {percent_progress:.1f}% -- Time remaining: {time_remaining} {unit}          ",
            end="\r",
            file=file_dest
        )

    return _print_progress


def now_real():
    return time.time()


def track_progress(
    measure: MeasureProgress,
    target: MetricProgress,
    interval_check: float,
    capture_maybe: Optional[CaptureProgress] = None
) -> None:
    """
    Tracks progress against a certain end condition of the simulation (for instance, a certain duration on the simulated
    clock), reporting this progress as the simulation chugs along. Stops the simulation once the target has been
    reached. By default, the progress is reported as printout on standard output, in a manner that works best for
    digital terminals.
    """

    def measure_to_target() -> MeasureComparison:
        return list(zip(measure(), target))

    def is_finished(progress: MeasureComparison) -> bool:
        return all(p >= t for p, t in progress)

    capture = capture_maybe or capture_print()
    rt_started = now_real()
    while True:
        advance(interval_check)

        rt_elapsed = now_real() - rt_started
        progress = measure_to_target()
        ratio_progress_min = min(m / t for m, t in progress)
        if ratio_progress_min == 0.0:
            rt_total_projected = inf
        else:
            rt_total_projected = rt_elapsed / ratio_progress_min
        capture(ratio_progress_min, rt_total_projected - rt_elapsed, progress)

        if is_finished(progress):
            stop()
            break


def _divide_round(dividend: Union[float, int], divider: Union[float, int]) -> int:
    return int(ceil(float(dividend) / divider))


def _display_time(seconds: float) -> Tuple[int, str]:
    if seconds == inf:
        delay = 1
        unit = "infinity"
    else:
        delay = _divide_round(seconds, 1)
        unit = "second"

        if delay > 90:
            delay = _divide_round(delay, 60)
            unit = "minute"

            if delay > 90:
                delay = _divide_round(delay, 60)
                unit = "hour"

                if delay > 36:
                    delay = _divide_round(delay, 24)
                    unit = "day"

        if delay > 1:
            unit += "s"

    return delay, unit



from setuptools import setup
# from subprocess import check_output

with open("README.md", "r") as file_long_description:
    long_description = file_long_description.read()

setup(
    name='greensim',
    version='2.1.0',
    packages=['greensim'],
    data_files=[('.', ['LICENSE'])],
    install_requires=['greenlet==0.4.14'],
    description='Discrete event simulation toolkit based on greenlets',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ElementAI/greensim",
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    )
)
